@@ -56,6 +56,7 @@ from tensorboard.backend import application
 from tensorboard.backend.event_processing import event_file_inspector as efi
 from tensorboard.plugins import base_plugin
 from tensorboard.plugins.core import core_plugin
+from tensorboard.util import argparse_util
 from tensorboard.util import tb_logging
 
 try:
@@ -68,6 +69,11 @@ except ImportError:
 
 logger = tb_logging.get_logger()
 
+# Default subcommand name. This is a user-facing CLI and should not change.
+_SERVE_SUBCOMMAND_NAME = 'serve'
+# Internal flag name used to store which subcommand was invoked.
+_SUBCOMMAND_FLAG = '__tensorboard_subcommand'
+
 
 def setup_environment():
   """Makes recommended modifications to the environment.
@@ -111,10 +117,13 @@ class TensorBoard(object):
     cache_key: As `manager.cache_key`; set by the configure() method.
   """
 
-  def __init__(self,
-               plugins=None,
-               assets_zip_provider=None,
-               server_class=None):
+  def __init__(
+      self,
+      plugins=None,
+      assets_zip_provider=None,
+      server_class=None,
+      subcommands=None,
+  ):
     """Creates new instance.
 
     Args:
@@ -133,9 +142,17 @@ class TensorBoard(object):
       assets_zip_provider = get_default_assets_zip_provider()
     if server_class is None:
       server_class = create_port_scanning_werkzeug_server
+    if subcommands is None:
+      subcommands = []
     self.plugin_loaders = [application.make_plugin_loader(p) for p in plugins]
     self.assets_zip_provider = assets_zip_provider
     self.server_class = server_class
+    self.subcommands = {}
+    for subcommand in subcommands:
+      name = subcommand.name()
+      if name in self.subcommands or name == _SERVE_SUBCOMMAND_NAME:
+        raise ValueError("Duplicate subcommand name: %r" % name)
+      self.subcommands[name] = subcommand
     self.flags = None
 
   def configure(self, argv=('',), **kwargs):
@@ -159,15 +176,48 @@ class TensorBoard(object):
     Raises:
       ValueError: If flag values are invalid.
     """
-    parser = argparse_flags.ArgumentParser(
+
+    base_parser = argparse_flags.ArgumentParser(
         prog='tensorboard',
         description=('TensorBoard is a suite of web applications for '
                      'inspecting and understanding your TensorFlow runs '
                      'and graphs. https://github.com/tensorflow/tensorboard '))
+    subparsers = base_parser.add_subparsers(
+        help="TensorBoard subcommand (defaults to %r)" % _SERVE_SUBCOMMAND_NAME)
+
+    serve_subparser = subparsers.add_parser(
+        _SERVE_SUBCOMMAND_NAME,
+        help='start local TensorBoard server (default subcommand)')
+    serve_subparser.set_defaults(**{_SUBCOMMAND_FLAG: _SERVE_SUBCOMMAND_NAME})
+
+    if len(argv) < 2 or argv[1].startswith('-'):
+      # This invocation, if valid, must not use any subcommands: we
+      # don't permit flags before the subcommand name.
+      serve_parser = base_parser
+    else:
+      # This invocation, if valid, must use a subcommand: we don't take
+      # any positional arguments to `serve`.
+      serve_parser = serve_subparser
+
+    for (name, subcommand) in six.iteritems(self.subcommands):
+      subparser = subparsers.add_parser(
+          name, help=subcommand.help(), description=subcommand.description())
+      subparser.set_defaults(**{_SUBCOMMAND_FLAG: name})
+      subcommand.define_flags(subparser)
+
     for loader in self.plugin_loaders:
-      loader.define_flags(parser)
+      loader.define_flags(serve_parser)
+
     arg0 = argv[0] if argv else ''
-    flags = parser.parse_args(argv[1:])  # Strip binary name from argv.
+
+    with argparse_util.allow_missing_subcommand():
+      flags = base_parser.parse_args(argv[1:])  # Strip binary name from argv.
+    if getattr(flags, _SUBCOMMAND_FLAG, None) is None:
+      # Manually assign default value rather than using `set_defaults`
+      # on the base parser to work around Python bug #9351 on old
+      # versions of `argparse`: <https://bugs.python.org/issue9351>
+      setattr(flags, _SUBCOMMAND_FLAG, _SERVE_SUBCOMMAND_NAME)
+
     self.cache_key = manager.cache_key(
         working_directory=os.getcwd(),
         arguments=argv[1:],
@@ -185,8 +235,9 @@ class TensorBoard(object):
       if not hasattr(flags, k):
         raise ValueError('Unknown TensorBoard flag: %s' % k)
       setattr(flags, k, v)
-    for loader in self.plugin_loaders:
-      loader.fix_flags(flags)
+    if getattr(flags, _SUBCOMMAND_FLAG) == _SERVE_SUBCOMMAND_NAME:
+      for loader in self.plugin_loaders:
+        loader.fix_flags(flags)
     self.flags = flags
     return [arg0]
 
@@ -208,14 +259,24 @@ class TensorBoard(object):
     :rtype: int
     """
     self._install_signal_handler(signal.SIGTERM, "SIGTERM")
-    if self.flags.inspect:
-      logger.info('Not bringing up TensorBoard, but inspecting event files.')
-      event_file = os.path.expanduser(self.flags.event_file)
-      efi.inspect(self.flags.logdir, event_file, self.flags.tag)
-      return 0
-    if self.flags.version_tb:
+    subcommand_name = getattr(self.flags, _SUBCOMMAND_FLAG)
+    if subcommand_name == _SERVE_SUBCOMMAND_NAME:
+      runner = self._run_serve_subcommand
+    else:
+      runner = self.subcommands[subcommand_name].run
+    return runner(self.flags) or 0
+
+  def _run_serve_subcommand(self, flags):
+    # TODO(#2801): Make `--version` a flag on only the base parser, not `serve`.
+    if flags.version_tb:
       print(version.VERSION)
       return 0
+    if flags.inspect:
+      # TODO(@wchargin): Convert `inspect` to a normal subcommand?
+      logger.info('Not bringing up TensorBoard, but inspecting event files.')
+      event_file = os.path.expanduser(flags.event_file)
+      efi.inspect(flags.logdir, event_file, flags.tag)
+      return 0
     try:
       server = self._make_server()
       server.print_serving_message()
@@ -300,6 +361,56 @@ class TensorBoard(object):
     return self.server_class(app, self.flags)
 
 
+@six.add_metaclass(ABCMeta)
+class TensorBoardSubcommand(object):
+  """Experimental private API for defining subcommands to tensorboard(1)."""
+
+  @abstractmethod
+  def name(self):
+    """Name of this subcommand, as specified on the command line.
+
+    This must be unique across all subcommands.
+
+    Returns:
+      A string.
+    """
+    pass
+
+  @abstractmethod
+  def define_flags(self, parser):
+    """Configure an argument parser for this subcommand.
+
+    Flags whose names start with two underscores (e.g., `__foo`) are
+    reserved for use by the runtime and must not be defined by
+    subcommands.
+
+    Args:
+      parser: An `argparse.ArgumentParser` scoped to this subcommand,
+        which this function should mutate.
+    """
+    pass
+
+  @abstractmethod
+  def run(self, flags):
+    """Execute this subcommand with user-provided flags.
+
+    Args:
+      flags: An `argparse.Namespace` object with all defined flags.
+
+    Returns:
+      An `int` exit code, or `None` as an alias for `0`.
+    """
+    pass
+
+  def help(self):
+    """Short, one-line help text to display on `tensorboard --help`."""
+    return None
+
+  def description(self):
+    """Description to display on `tensorboard SUBCOMMAND --help`."""
+    return None
+
+
 @six.add_metaclass(ABCMeta)
 class TensorBoardServer(object):
   """Class for customizing TensorBoard WSGI app serving."""
